@@ -27,8 +27,13 @@
 
 #include <unistd.h>
 
+#ifdef HAVE_XEXTPROTO_71
+#include <X11/extensions/dpmsconst.h>
+#else
 #define DPMS_SERVER
 #include <X11/extensions/dpms.h>
+#endif
+
 #include <xf86_OSproc.h>
 
 #include "g80_type.h"
@@ -77,7 +82,7 @@ G80DacDPMSSet(xf86OutputPtr output, int mode)
     pNv->reg[(0x0061A004+off)/4] = tmp;
 }
 
-Bool
+static Bool
 G80DacModeFixup(xf86OutputPtr output, DisplayModePtr mode,
                 DisplayModePtr adjusted_mode)
 {
